#ifndef COMBINED_DATA_CONFLICT_TABLE_H_INCLUDED_
#define COMBINED_DATA_CONFLICT_TABLE_H_INCLUDED_

#include<string>
#include<utility>
#include<vector>

#include "data_conflict_table.h"

#include "segment_graph.h" // needed for VertexDescriptor
#include "mapped_variable.h" // needed for MappedVariable

/**
 * INFORMATION / EFFICIENCY
 * I tried out boost::numerics::ublas::matrix
 * instead of a self implemented matrix.
 * Unfortunately, boost is slower by factor 4 at least.
 */

namespace risc {

namespace sg {

class CombinedDataConflictTable {

public:

  /**
   * \fn    CombinedDataConflictTable(
            risc::sg::DataConflictTable &dct,
            PathInstanceMapper *path_instance_mapper)
   * \brief Creates the combined data conflict table.
   *        The maximum number of steps will be calculated automatically
   */
  CombinedDataConflictTable(
      risc::sg::DataConflictTable &dct,
      PathInstanceMapper *path_instance_mapper);

  ~CombinedDataConflictTable();

  /**
   * \fn    get_max_instances(int segment_id)
   * \brief This functions returns how many instances of a given segment exist
   */
  int get_max_instances(int segment_id);

  /**
   * \fn    index_to_segment_and_instance_id(int index)
   * \brief This function translates a given index from the conflict table
   *        into the corresponding segment id and instance id. The first
   *        value is the segment id and the second value is the instance id.
   */
  std::pair<int, int> index_to_segment_and_instance_id(int index);

  /**
   * \fn    segment_and_instance_id_to_index(int segment_id, int instance_id)
   * \brief This function translates a given segment id and instance id
   *        to a index of the conflict table
   */
  int segment_and_instance_id_to_index(int segment_id, int instance_id);

  /**
   * \fn    print_conflict_table
   * \brief This function prints the combined conflict table in the terminal
   */
  void print_conflict_table();

  /**
   * \fn    print_conflict_table(bool *table)
   * \brief This function prints the table table in the terminal
   */
  void print_conflict_table(bool *table);

  /**
   * \fn    print_conflict_table(std::string filename)
   * \brief This function prints the combined conflict table
   *        in the file filename
   */
  void print_conflict_table(std::string filename);

  /**
   * \fn    print_lookup_table
   * \brief This function prints the table instance_id_to_index_lookup_table
   *        in the terminal
   */
  void print_lookup_table();

  /**
   * \fn    print_lookup_table(std::string filename)
   * \brief This function prints the table instance_id_to_index_lookup_table
   *        in the file filename
   */
  void print_lookup_table(std::string filename);

  /**
   * \fn    max_prediction_steps
   * \brief the number of prediction steps until no more changes will happen
   */
  int max_prediction_steps();

  /**
   * \brief The global path instance mapper
   */
  PathInstanceMapper *path_instance_mapper_;

  /**
   * \brief This lookup table translates segment ids into vertex descriptors.
   *        At position i is the segment with the id of
   *        instance_id_to_index_lookup_[0][i]
   */
  VertexDescriptor *vertex_lookup_;

  /**
   * \brief This variable represents the max number of module instances in the
   *        design. Also, this is the number of rows in the
   *        instance_id_to_index_lookup_ table.
   */
  int max_instances_;

  /**
   * \brief This variable defines how many segments exist in the segment
   *        graph. Also, this is the number of columns in the
   *        instance_id_to_index_lookup_ table.
   */
  int number_of_segments_;

  /**
   * \brief This variable defines the size of the conflict table.
   *        The size is quadratic.
   */
  int size_of_conflict_table_;

  /**
   * \brief This table should be used a lookup table:
   *        (segment id, instance id) -> index of conflict table
   */
  int **instance_id_to_index_lookup_;

  /**
   * \brief The conflict table
   */
  int *combined_table_;

  /**
   * \brief number of prediction steps until no more changes happen
   */
  int prediction_steps_;

private:

  /**
   * \fn    compute_combined_table
   * \brief This function computes the combined table for 'prediction_steps_'
   *        prediction steps
   */
  void compute_combined_table();

  /**
   * \brief The data conflict table
   */
  risc::sg::DataConflictTable &dct_;
};

} // end of namespace sg

} // end of namespace risc

#endif /* COMBINED_DATA_CONFLICT_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
