#ifndef CONFLICT_TABLE_H_INCLUDED_
#define CONFLICT_TABLE_H_INCLUDED_

#include<utility>

#include "rose.h"

#include "segment_graph.h" // needed for VertexDescriptor
#include "mapped_variable.h" // needed for MappedVariable

namespace risc {

class PathInstanceMapper;

namespace sg {

class ConflictTable {

public:

  /**
   * \brief Default constructor
   */
  ConflictTable(
      SegmentGraph &segment_graph,
      PathInstanceMapper *path_instance_mapper,
      bool debugging = false);

  /**
   * \brief Destructor
   */
  ~ConflictTable();

  /**
   * \brief This lookup table translates segment ids into vertex descriptors.
   *        At position i is the segment with the id of
   *        instance_id_to_index_lookup_[0][i]
   */
  VertexDescriptor *vertex_lookup_;

  /**
   * \brief This variable defines the size of the conflict table.
   *        The size is quadratic.
   */
  int size_of_conflict_table_;

  /**
   * \brief Maximum number of module/channel instances
   *        Only instances with at least one simulation thread are considered
   */
  int max_instances_with_thread_;

  /**
   * \brief This variable represents the max number of module instances in the
   *        design. Also, this is the number of rows in the
   *        instance_id_to_index_lookup_ table.
   */
  int max_instances_;

  /**
   * \brief This variable defines how many segments exist in the segment
   *        graph. Also, this is the number of columns in the
   *        instance_id_to_index_lookup_ table.
   */
  int number_of_segments_;

  /**
   * \brief This table should be used a lookup table:
   *        (segment id, instance id) -> index of conflict table
   */
  int **instance_id_to_index_lookup_;

  /**
   * \brief The conflict table
   */
  bool **conflict_table_;

  /**
   * \fn       print_conflict_table(std::string filename)
   * \brief    Generates a html file of the conflict table
   */
  void print_conflict_table(std::string filename);

  /**
   * \fn       print_lookup_table(std::string filename)
   * \brief    Generates a html file of the lookup table
   */
  void print_lookup_table(std::string filename);

  /**
   * \fn       segment_and_instance_id_to_index(int segment_id, int instance_id)
   * \brief    This function translates a given segment id and instance id
   *           to a index of the conflict table
   */
  int segment_and_instance_id_to_index(int segment_id, int instance_id);

  /**
   * \fn       index_to_segment_and_instance_id(int index)
   * \brief    This function translates a given index from the conflict table
   *           into the corresponding segment id and instance id. The first
   *           value is the segment id and the second value is the instance id.
   */
  std::pair<int, int> index_to_segment_and_instance_id(int index);

  /**
   * \fn       has_conflict(int segment_id_1, int instance_id_1, int segment_id_2,
      			int instance_id_2)
   * \brief    This function looks up if there is a conflict between segment
   *           1 with instance id 1 and segment 2 with instance id 2
   */
  bool has_conflict(int segment_id_1, int instance_id_1, int segment_id_2,
      int instance_id_2);

  /**
   * \fn       print_instance_id_to_index_lookup_table
   * \brief    This function prints the table instance_id_to_index_lookup_table
   *           in the terminal
   */
  void print_instance_id_to_index_lookup_table();

  /**
   * \fn       set_conflict(int segment_id_1, int instance_id_1,
      int segment_id_2, int instance_id_2)
   * \brief    This function sets a conflict in the conflict table
   */
  void set_conflict(int segment_id_1, int instance_id_1,
      int segment_id_2, int instance_id_2);

  /**
   * \fn       determine_conflict_table
   * \brief    This function determines the conflict table.
   *           The function is overloaded for the data and event conflict table.
   */
  virtual void determine_conflict_table();

  /**
   * \fn       segment_id_to_vertex_descriptor(int id)
   * \brief    This function translates the given segment id to the
   *           corresponding vertex descriptor
   */
  VertexDescriptor segment_id_to_vertex_descriptor(int id);

  /**
   * \fn       get_mapped_variable(SgVariableSymbol *variable, int instance_id,
      bool dynamic_analysis)
   * \brief    Determines the mapped variable
   */
  MappedVariableList
  get_mapped_variable(SgVariableSymbol *variable, int instance_id,
      bool dynamic_analysis);

  /**
   * \fn       get_mapped_variable_statically(SgVariableSymbol *variable, int instance_id)
   * \brief    This function determines for a given variable the mapped
   *           variable. If variable is a reference, the mapped variable will
   *           be looked up. Otherwise, the original symbol and instance id
   *           is returned.
   */
  MappedVariableList
  get_mapped_variable_statically(SgVariableSymbol *variable, int instance_id);

  /**
   * \fn       get_mapped_variable_dynamically(SgVariableSymbol *variable, int instance_id)
   * \brief    This function determines the mapped address of a reference
   *           during the pre simulation.
   */
  MappedVariableList
  get_mapped_variable_dynamically(SgVariableSymbol *variable, int instance_id);

  /**
   * \fn       get_max_instances(int segment_id)
   * \brief    This functions returns how many instances of a given segment
   *           exist
   */
  int get_max_instances(int segment_id);

  /**
   * \fn             get_channel_id(const SymbolWithPath &swp, int instance,
      bool dynamic_analysis)
   * \brief          Determines the channel instance id for a given variable
   *                 declared in a channel
   * \param swp      Symbol with the associated port call path
   * \param instance instance id of the hosting module
   * \param dynamic_analysis if the analysis is dynamic_analysis
   */
  int
  get_channel_id(const SymbolWithPath &swp, int instance,
      bool dynamic_analysis);

  /**
   * \fn             get_channel_id_statically(const SymbolWithPath &swp, int instance)
   * \brief          Determines the channel instance id for a given variable
   *                 declared in a channel
   * \param swp      Symbol with the associated port call path
   * \param instance instance id of the hosting module
   */
  int get_channel_id_statically(const SymbolWithPath &swp, int instance);

  /**
   * \fn             get_channel_id_dynamically(const SymbolWithPath &swp, int instance)
   * \brief          Determines the channel instance id for a given variable
   *                 declared in a channel
   * \param swp      Symbol with the associated port call path
   * \param instance instance id of the hosting module
   */
  int get_channel_id_dynamically(const SymbolWithPath &swp, int instance);

  PathInstanceMapper *path_instance_mapper_;

  SegmentGraph &segment_graph_;

  // The following part is for debugging purposes
  struct ConflictInformation {
    std::set<Conflict> conflict_variables_;
    bool contains_sc_stop_;
    ConflictInformation():contains_sc_stop_(false){}
  };

  ConflictInformation **conflict_info_;

  bool debugging_;
};

}; // end of namesapce sg

}; // end of namespace risc

#endif /* CONFLICT_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
