#include "event_conflict_table.h"

#include <list>
#include <unordered_set>
#include "mapped_variable.h"
#include "segment.h"
#include "segment_graph.h"
#include "../tools/variable_helper.h"
#include "../tools/class_is_derived_from.h"
#include "../tools/misc.h"
#include "../tools/systemc_definitions.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/instance.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/module.h"
#include "../internal_representation/path_instance_mapper.h"

std::list<risc::sg::MappedVariable>
risc::sg::EventConflictTable::get_all_list_events_dynamically(
    SgVariableSymbol *sc_event_list, int instance_id)
{
  // here we are getting the variable of the event list.
  // Just in case that the event list is a reference
  MappedVariableList mapped_event_list
    = get_mapped_variable_dynamically(sc_event_list, instance_id);

  // getting the variable declaration from the event list
  SgVariableDeclaration *variable
    = isSgVariableDeclaration(
          mapped_event_list.front().symbol_->get_declaration()
                                               ->get_declaration());

  SgClassDefinition *enclosing_class_def_of_event_list
    = SageInterface::getEnclosingClassDefinition(variable);

  // getting the name of the variable
  std::string host_module_type_name
    = risc::tools::get_mangled_name(enclosing_class_def_of_event_list);

  if (host_module_type_name.substr(0, 6) == "class_") {
    host_module_type_name.erase(0, 6);
  }

  std::string event_list_name = SageInterface::get_name(variable);
  event_list_name.erase(0, 22);

  // here we are getting the DynamicObject of the event list in the tree
  int counter = 0;
  DynamicObject *mapped_variable_address
    = get_dynamic_object_of_variable(
          path_instance_mapper_->get_dynamic_tree_root(),
          host_module_type_name,
          mapped_event_list.front().instance_id_,
          event_list_name,
          counter);

  DynamicEventList *dynamic_event_list
    = dynamic_cast<DynamicEventList*>(mapped_variable_address);

  std::list<risc::sg::MappedVariable> list_events;

  for(std::list<int*>::iterator
      iter  = dynamic_event_list->events_.begin();
      iter != dynamic_event_list->events_.end();
      iter++) {

    // Searching for the host module and the variable name of the variable
    // address
    std::pair<DynamicModule*, std::string> host_module_and_variable
      = get_host_module_and_name_of_variable(
          path_instance_mapper_->get_dynamic_tree_root(),
          *iter);

    // If we are reaching this point, it is a member variable
    // So, next we determine the instance id of the hosting module of the
    // variable
    int mapped_instance_id = 0;

    bool result = get_instance_id_of_hosting_module(
                      path_instance_mapper_->get_dynamic_tree_root(),
                      *iter,
                      host_module_and_variable.first->name_,
                      mapped_instance_id);

    // At this point, we have to find the associated module definition
    // in the internal representation. We can do this by comparing the
    // modules types (mangled names)
    risc::Module *current_module
      = risc::find_module_definition(host_module_and_variable.first->name_);

    SgVariableSymbol *mapped_symbol = NULL;

    // After having the module, we have to search for the mapped variable in the
    // class. So, we are iterating over all the declarations in the class and
    // compare all the variables names with name of the mapped variable
    for(SgDeclarationStatementPtrList::iterator
        member_iter  = current_module->get_ast_node()->get_members().begin();
        member_iter != current_module->get_ast_node()->get_members().end();
        member_iter++) {

      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
      if(var_decl) {

        std::string temp_variable_name(SageInterface::get_name(var_decl));

        // deleting __variable_declaration__ out of the string
        temp_variable_name.erase(0, 22);

        if(temp_variable_name == host_module_and_variable.second) {
          mapped_symbol = isSgVariableSymbol(var_decl->get_decl_item(0)
              ->search_for_symbol_from_symbol_table());
        }
      }
    }

    // pushing the mapped event into the event list
    list_events.push_back(MappedVariable(mapped_symbol, mapped_instance_id));
  }

  return list_events;
}

// RD: this searches for event lists defined in "before_end_of_elaboration"
// RD: (and nowhere else, i.e. not in the constructor)
// RD: and doesn't distinguish between event and- and or-lists
// RD: (OPEN QUESTION: does this repeat work done in the IR? if so, why?)
std::list<risc::sg::MappedVariable>
risc::sg::EventConflictTable::get_all_list_events_statically(
    SgSymbol *sc_event_list, int instance_id, PortCallPath pcp)
{
  std::list<MappedVariable> result;

  SgClassDefinition *class_def
    = SageInterface::getEnclosingClassDefinition(sc_event_list);

  std::vector<SgNode*> func_defs
    = NodeQuery::querySubTree(class_def, V_SgFunctionDefinition);

  SgFunctionDefinition *before_end_of_elaboration = NULL;

  for(std::vector<SgNode*>::iterator
      iter  = func_defs.begin();
      iter != func_defs.end();
      iter++) {

    SgFunctionDefinition *func_def = isSgFunctionDefinition(*iter);
    if(SageInterface::get_name(func_def) == "before_end_of_elaboration") {
      before_end_of_elaboration = func_def;
      break; // we found it, we can stop searching now (RD, 06/27/18)
    }
  }
  if (!before_end_of_elaboration) { // RD: not supplied in this sc_module
    return result;  // return empty list
  }

  for(SgStatementPtrList::iterator
      iter  = before_end_of_elaboration->get_body()->get_statements().begin();
      iter != before_end_of_elaboration->get_body()->get_statements().end();
      iter++) {

    // RD: Tim searches now for event list assignments, e.g. el &= e1;
    // RD: this is a pretty ad-hoc implementation...
    SgExprStatement *expr_stmt = isSgExprStatement(*iter);

    if(!expr_stmt) {
      continue;
    }

    SgExpression *expr = isSgExpression(expr_stmt->get_expression());

    if(!expr) {
      continue;
    }

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expr);
    if(func_call_exp) {

    // bug fix: Tim wrote this...
    //if((SageInterface::get_name(func_call_exp).find("&=")) ||
    //    (SageInterface::get_name(func_call_exp).find("|=")))
    // bug fix: ...but surely meant this (RD, 06/27/18)
      if((SageInterface::get_name(func_call_exp).find("&=")!=std::string::npos) ||
          (SageInterface::get_name(func_call_exp).find("|=")!=std::string::npos)) {

        SgVarRefExp *assigned_var =
          isSgVarRefExp(
              isSgArrowExp(
                  isSgDotExp(
                      func_call_exp->get_function())
                  ->get_lhs_operand())
              ->get_rhs_operand());

        if(assigned_var) {

          if(assigned_var->get_symbol() == sc_event_list) {

            SgExprListExp *expr_list = func_call_exp->get_args();

// OLD: Tim's code considers only one plain event (for overloaded assignment RHS):
//      Example: elist |= e1;
//
//          SgVarRefExp *var
//            = isSgVarRefExp(
//                  isSgArrowExp(
//                      expr_list->get_expressions().at(0))
//                  ->get_rhs_operand());
//           MappedVariableList mapped_event
//            = get_mapped_variable_statically(var->get_symbol(), instance_id);
//
//          if(tools::is_channel_variable(mapped_event.front().symbol_)) {
//            SymbolWithPath temp(mapped_event.front().symbol_, pcp);
//            mapped_event.front().instance_id_
//              = get_channel_id(temp, mapped_event.front().instance_id_,
//                  false);
//          }
//          result.push_back(mapped_event.front());
//
// NEW: let ROSE do a traversal and add all events found (RD, 06/27/18)
//      Example: elist |= e1 | e2 | (e3 | e4);

            std::vector<SgNode*> var_ref =
              NodeQuery::querySubTree(expr_list->get_expressions().at(0), V_SgVarRefExp);
            for(std::vector<SgNode*>::iterator
                iter  = var_ref.begin();
                iter != var_ref.end();
                iter++) {
              SgVarRefExp *var
                = isSgVarRefExp(*iter);
              //printf("Event %s\n", var->get_symbol()->get_name().str());

              MappedVariableList mapped_event
                = get_mapped_variable_statically(var->get_symbol(), instance_id);

              if(tools::is_channel_variable(mapped_event.front().symbol_)) {
                SymbolWithPath temp(mapped_event.front().symbol_, pcp);
                mapped_event.front().instance_id_
                  = get_channel_id(temp, mapped_event.front().instance_id_,
                      false);
              }
              result.push_back(mapped_event.front());
            }
// END OF NEW CODE (RD, 06/27/18)
          }
        }
      }
    }
  }

  return result;
}

risc::sg::EventConflictTable::EventConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper *path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  determine_conflict_table();
}

void
risc::sg::EventConflictTable::determine_conflict_table()
{
  typedef boost::graph_traits<Graph>::vertex_iterator
                              VertexIterator; // (Tim's segment iterator)
  std::pair<VertexIterator, VertexIterator> vp;

  bool dynamic_analysis = path_instance_mapper_->perform_dynamic_analysis();

// RD: STEP 1, we gather all the events each segment is waiting for

  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second;
      ++vp.first) {
    Segment &ColSegment = segment_graph_.graph_[*vp.first];

    // RD: we keep this list stored with the segment for later reuse (NEW)
    std::set<SymbolWithPath> &wait_events = ColSegment.wait_events_;
    //Commented off by ZC
    //Not true for loaded PSG
    //assert(wait_events.empty());  // this hasn't been computed before

    const SgNode *boundary = ColSegment.boundary_node_;
    const SgStatement *stmt = isSgStatement(boundary);
    const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

    //no longer needed, ZC
    /*
    for(std::vector<SgVariableSymbol*>::iterator // ZC's PSG support
        iter = ColSegment.wait_event_symbols.begin();
        iter!= ColSegment.wait_event_symbols.end();
        iter++){
      SymbolWithPath wait_event((*iter), ColSegment.pcp_); // RD+ZC: ....BUG!
      wait_events.push_back(wait_event); // gather any events from imported PSGs
    }
    */
    // Here, we have to check what kind of boundary statement it is. If it is a
    // wait function call, we have to analyze the arguments. So, if we have
    // exactly one argument it could be an event or an event list. We will check
    // if the symbol is of the type event.
    if(expr_stmt) {
      const SgFunctionCallExp *func_call
        = isSgFunctionCallExp(expr_stmt->get_expression());
      if(func_call) {
        if(SageInterface::get_name(
               func_call->getAssociatedFunctionDeclaration()) == "wait") {

          // At this point, we know that the boundary statement is a function
          // call of wait.
          // Next we have to figure out what are the arguments

          if(func_call->get_args()->get_expressions().size() == 0) {

            // In case we have no argument, it's static sensitivity
            // (RD, 06/27/18: Tim called this incorrectly "an immediate notification")
            // Example: wait();
            // TS: We cannot support this one (v0.4.2 and earlier)
            // RD: RISC IR already contains the static sensitivity for every thread,
            //     we only need to find out which thread calls this wait!
            //     NEW v0.4.3: Originating thread is stored in segments where it is known.
            Thread *OriginatingThread = ColSegment.originating_thread_;
            if (OriginatingThread) {  // an originating thread exists
              if (OriginatingThread == (Thread*)-1) { // RD hack: multiple threads reach this
                std::cout << std::endl
                          << "Error: Multiple threads! Cannot determine sensitivity of wait()"
                          << std::endl
                          << "       in "
                          << func_call->get_file_info()->get_filename()
                          << " line "
                          << func_call->get_file_info()->get_line()
                          << ". Aborting."
                          << std::endl;
                exit(10);
              }
              //OriginatingThread->print_event_list();
              for(EventVectorIter         // add event sensitivity
                  iter = OriginatingThread->static_event_sensitivity_list_.begin();
                  iter != OriginatingThread->static_event_sensitivity_list_.end();
                  iter++) {
                SymbolWithPath wait_event((*iter)->get_symbol(), ColSegment.pcp_);
                // ZC, no longer needed
                // if(PSG_MODE){
                //   ColSegment.wait_event_symbols.push_back(wait_event.symbol_);
                // }
                wait_events.insert(wait_event);
              }
              for(PortVectorIter          // add port sensitivity
                  iter = OriginatingThread->static_port_sensitivity_list_.begin();
                  iter != OriginatingThread->static_port_sensitivity_list_.end();
                  iter++) {
                SymbolWithPath wait_event((*iter)->get_symbol(), ColSegment.pcp_);
                // ZC, no longer needed
                // if(PSG_MODE){
                //   ColSegment.wait_event_symbols.push_back(wait_event.symbol_);
                // }
                wait_events.insert(wait_event);
              }
            } else {                  // no originating thread is known, stop!
              std::cout << std::endl
                        << "Error: Unknown thread! Cannot determine sensitivity of wait()"
                        << std::endl
                        << "       in "
                        << func_call->get_file_info()->get_filename()
                        << " line "
                        << func_call->get_file_info()->get_line()
                        << ". Aborting."
                        << std::endl;
              exit(10);
            }
          } else if(func_call->get_args()->get_expressions().size() == 1) {

            // A sc_zero_time will wait for one delta count. We don't have to
            // set the wait symbol
            // Example: wait(SC_ZERO_TIME);
            if(func_call->get_args()->get_expressions().at(0)->unparseToString()
                == "SC_ZERO_TIME") {
              // no event to store
            } 
            //DM 05/20/2019 for sc_method event analysis
            else if(func_call->get_args()->get_expressions().at(0)->unparseToString()
                == "((this) -> simcontext())") {
            Thread *OriginatingThread = ColSegment.originating_thread_;
            if (OriginatingThread) {  // an originating thread exists
              if (OriginatingThread == (Thread*)-1) { // RD hack: multiple threads reach this
                std::cout << std::endl
                          << "Error: Multiple threads! Cannot determine sensitivity of wait()"
                          << std::endl
                          << "       in "
                          << func_call->get_file_info()->get_filename()
                          << " line "
                          << func_call->get_file_info()->get_line()
                          << ". Aborting."
                          << std::endl;
                exit(10);
              }
              //OriginatingThread->print_event_list();
              for(EventVectorIter         // add event sensitivity
                  iter = OriginatingThread->static_event_sensitivity_list_.begin();
                  iter != OriginatingThread->static_event_sensitivity_list_.end();
                  iter++) {
                SymbolWithPath wait_event((*iter)->get_symbol(), ColSegment.pcp_);
                // ZC, no longer needed
                // if(PSG_MODE){
                //   ColSegment.wait_event_symbols.push_back(wait_event.symbol_);
                // }
                wait_events.insert(wait_event);
              }
              for(PortVectorIter          // add port sensitivity
                  iter = OriginatingThread->static_port_sensitivity_list_.begin();
                  iter != OriginatingThread->static_port_sensitivity_list_.end();
                  iter++) {
                SymbolWithPath wait_event((*iter)->get_symbol(), ColSegment.pcp_);
                // ZC, no longer needed
                // if(PSG_MODE){
                //   ColSegment.wait_event_symbols.push_back(wait_event.symbol_);
                // }
                wait_events.insert(wait_event);
              }
              for(std::list<NodeWithPath>::iterator expr_iter = ColSegment.expressions_.begin();
                expr_iter != ColSegment.expressions_.end();
                expr_iter++) {
                const SgFunctionCallExp *func_call
                  = isSgFunctionCallExp(expr_iter->node_);
                if(func_call && func_call->getAssociatedFunctionDeclaration() && 
                   SageInterface::get_name(func_call->getAssociatedFunctionDeclaration()) == "next_trigger") {
           
                   if(func_call->get_args()->get_expressions().size() == 0) {
                    //do nothing
                   }
                   else if(func_call->get_args()->get_expressions().size() == 1) {
                      //add event
                     std::vector<SgNode*> var_ref =
                     NodeQuery::querySubTree(func_call->get_args()->get_expressions().at(0),
                                        V_SgVarRefExp);

                     for(std::vector<SgNode*>::iterator
                      iter  = var_ref.begin();
                      iter != var_ref.end();
                      iter++) {
                      SgVarRefExp *wait_var = isSgVarRefExp(*iter);
                      //printf("wait_var %s\n", wait_var->get_symbol()->get_name().str());

                      if (isSgPointerType(wait_var->get_type()->stripTypedefsAndModifiers())) {

                       std::cout << std::endl
                            << "Error: No support of event pointers." << std::endl
                            << "       See line "
                            << func_call->get_file_info()->get_line()
                            << " in file "
                            << func_call->get_file_info()->get_filename()
                            << std::endl;
                       exit(1);

                      } else {
                      // we now have a variable or reference of type
                      // sc_event, sc_event_and_list, or sc_event_or_list
                       // RD TODO....: add/handle sc_signal and family here!
                      //....assert
                        if( risc::tools::is_sc_time(wait_var) )
                        {//do nothing
                        continue;
                        }
                        else if (! (  (risc::tools::is_sc_event(wait_var))
                          ||(risc::tools::is_sc_event_and_list(wait_var))
                          ||(risc::tools::is_sc_event_or_list(wait_var))
                         )) {
                          printf("\nASSERTION WARNING(wait): Neither sc_event, nor sc_event_list!\n");
                         }
                         SymbolWithPath wait_event(wait_var->get_symbol(), ColSegment.pcp_);
                         // ZC, no longer needed
                         // if(PSG_MODE){
                         //   ColSegment.wait_event_symbols.push_back(wait_event.symbol_);
                         // }
                         wait_events.insert(wait_event);
                      }
                    } 

                  }
                  else if(func_call->get_args()->get_expressions().size() == 2) {
                    //do nothing
                  }
                  else if(func_call->get_args()->get_expressions().size() == 3) {
                  }
          
                }
              }


            } else {                  // no originating thread is known, stop!
              std::cout << std::endl
                        << "Error: Unknown thread! Cannot determine sensitivity of wait()"
                        << std::endl
                        << "       in "
                        << func_call->get_file_info()->get_filename()
                        << " line "
                        << func_call->get_file_info()->get_line()
                        << ". Aborting."
                        << std::endl;
              exit(10);
            }
 
            }
            //end DM
            else {

              // Here we need to identify all events for which the wait function
              // call is waiting; there is only one argument, an event expression,
              // which might be a simple event or a complex event expression
              // Example: wait(e1|e2|some_event_list);
              // NEW: let ROSE do a traversal and add all events found (RD, 06/27/18)
              std::vector<SgNode*> var_ref =
                NodeQuery::querySubTree(func_call->get_args()->get_expressions().at(0),
                                        V_SgVarRefExp);
              for(std::vector<SgNode*>::iterator
                    iter  = var_ref.begin();
                    iter != var_ref.end();
                    iter++) {
                SgVarRefExp *wait_var = isSgVarRefExp(*iter);
                //printf("wait_var %s\n", wait_var->get_symbol()->get_name().str());

                if (isSgPointerType(wait_var->get_type()->stripTypedefsAndModifiers())) {

                  std::cout << std::endl
                            << "Error: No support of event pointers." << std::endl
                            << "       See line "
                            << func_call->get_file_info()->get_line()
                            << " in file "
                            << func_call->get_file_info()->get_filename()
                            << std::endl;
                  exit(1);

                } else {
                  // we now have a variable or reference of type
                  // sc_event, sc_event_and_list, or sc_event_or_list
                  // RD TODO....: add/handle sc_signal and family here!
                  //....assert
                  if( risc::tools::is_sc_time(wait_var) )
                  {//do nothing
                    continue;
                  }
                  else if (! (  (risc::tools::is_sc_event(wait_var))
                         ||(risc::tools::is_sc_event_and_list(wait_var))
                         ||(risc::tools::is_sc_event_or_list(wait_var))
                        )) {
                    printf("\nASSERTION WARNING(wait): Neither sc_event, nor sc_event_list!\n");
                  }
                  SymbolWithPath wait_event(wait_var->get_symbol(), ColSegment.pcp_);
                  // ZC, no longer needed
                  // if(PSG_MODE){
                  //   ColSegment.wait_event_symbols.push_back(wait_event.symbol_);
                  // }
                  wait_events.insert(wait_event);
                }
              } // end of for (RD, 06/27/18)
            }
          }
        }
      }
    }
#if 0  // RD: PRINT wait_events FOR DEBUGGING
    printf("DEBUG: wait_events(Seg%d)", ColSegment.id_);
    for(std::list<SymbolWithPath>::iterator
        iter  = wait_events.begin();
        iter != wait_events.end();
        *iter++) {
      printf(", (%s,{%d})", iter->symbol_->get_name().str(), iter->pcp_.path_.size());
    }
    printf("\n");
#endif // RD: PRINT wait_events FOR DEBUGGING
  }

// RD: STEP 2, we gather all the events notified in each segment
  std::unordered_set<SgVariableSymbol*> global_events;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second;
      ++vp.first) 
  {
      Segment &RowSegment = segment_graph_.graph_[*vp.first];

      // RD: we keep this list stored with the segment for later reuse (NEW)
      std::set<SymbolWithPath> &notify_events = RowSegment.notify_events_;
      //ZC july 25 2018, commented off next line,
      //because for loaded psgs, it is not true
      //assert(notify_events.empty());  // this hasn't been computed before

      // We are iterating over all expressions for each segment. For an
      // individual segment we have to check if there are any notifications of
      // events
      for(Segment::Expressions::iterator
          iter  = RowSegment.expressions_.begin();
          iter != RowSegment.expressions_.end();
          iter++) 
      {
        SgFunctionCallExp *func_call = isSgFunctionCallExp(iter->node_);
//      if (func_call)
//      { printf("RDDEBUG: %s\n", func_call->unparseToString().c_str()); }
        SgDotExp *dot_expr = NULL;
        SgFunctionDeclaration *func_decl = NULL;
        SgArrowExp *arrow_expr = NULL;
        SgVarRefExp *notify_event = NULL;
        bool is_global = false;
        if(  (func_call)
           &&(dot_expr = isSgDotExp(func_call->get_function()))
           &&(func_decl = func_call->getAssociatedFunctionDeclaration())
           &&(SageInterface::get_name(func_decl) == "notify")) {
          // at this point, we have a function call "someobject.notify(someargumentlist)" (RD)
          if((arrow_expr = isSgArrowExp(dot_expr->get_lhs_operand()))) {
            // here someobject is "exprleft->exprright", so TS assumes "this->e"
            notify_event = isSgVarRefExp(arrow_expr->get_rhs_operand());
          } else {
            // here someobject is "exprleft", so we assume "e" (bug fix 02/11/19, RD)
            notify_event = isSgVarRefExp(dot_expr->get_lhs_operand());
            is_global = true;
          }
          if(  (notify_event)
             &&(risc::tools::is_type_sc_event( // (fix for bug20180717a, 07/23/18, RD)
                    notify_event->get_type()->findBaseType()))) {
              // At this point, we have identified a call to (this->e).notify
              // or a call e.notify (bug fix 02/11/19, RD)

              if(func_call->get_args()->get_expressions().size() == 0) {
                // In case we have an immediate notification.
                // Example: e.notify();
                // We cannot support this one
                std::cout << std::endl
                          << "Error: No support of immediate event notification."
                          << std::endl
                          << "       See line "
                          << func_call->get_file_info()->get_line()
                          << " in "
                          << func_call->get_file_info()->get_filename()
                          << std::endl;
                exit(1);
              }
              if(func_call->get_args()->get_expressions().size() == 1) {
                // At this point we know that it is a event notification with one argument
                // Example: e.notify(SC_ZERO_TIME);
                if(func_call->get_args()->get_expressions().at(0)->unparseToString()
                                  != "SC_ZERO_TIME") {  // RD: ensure we wait for SC_ZERO_TIME
                  std::cout << std::endl
                            << "Error: Expected notify(SC_ZERO_TIME)."
                            << std::endl
                            << "       See line "
                            << func_call->get_file_info()->get_line()
                            << " in file "
                            << func_call->get_file_info()->get_filename()
                            << std::endl;
                  exit(1);
                }
                assert(NULL==isSgPointerType( // cannot be a pointer type here
                            notify_event->get_type()->stripTypedefsAndModifiers()));
                // RD TODO....: add/handle sc_signal and family here!
                if(risc::tools::is_sc_event(notify_event)) {
                  SymbolWithPath notifying_event(notify_event->get_symbol(),
                                                  iter->pcp_); // (not RowSegment.pcp_!)
                  // ZC, no longer needed
                  // if(PSG_MODE){
                  //   RowSegment.notify_event_symbols.push_back(notifying_event.symbol_);
                  // }
                  if(is_global) global_events.insert(notify_event->get_symbol());
                  notify_events.insert(notifying_event);
                }
              } else {
                // At this point we have a notify-call with 2 or more arguments
                // Example: e.notify(42, SC_MS);
                std::cout << std::endl
                          << "Error: Expected notify(SC_ZERO_TIME)."
                          << std::endl
                          << "       See line "
                          << func_call->get_file_info()->get_line()
                          << " in file "
                          << func_call->get_file_info()->get_filename()
                          << std::endl;
                exit(1);
              }
          }
        }
      }
       // ZC, no longer needed
       /*
       } // rof(Expressions)
       for(std::vector<SgVariableSymbol*>::iterator    // RD: ZC's PSG support
          iter = RowSegment.notify_event_symbols.begin();
          iter!= RowSegment.notify_event_symbols.end();
          iter++){
        SymbolWithPath notify_event((*iter), RowSegment.pcp_);
        notify_events.push_back(notify_event);
       }
       */
#if 0  // RD: PRINT notify_events FOR DEBUGGING
       printf("DEBUG: notify_events(Seg%d)", RowSegment.id_);
       for(std::list<SymbolWithPath>::iterator
          iter  = notify_events.begin();
          iter != notify_events.end();
          *iter++) {
         printf(", (%s,{%d})", iter->symbol_->get_name().str(), iter->pcp_.path_.size());
       }
       printf("\n");
#endif // RD: PRINT notify_events FOR DEBUGGING
  }

// RD: STEP 3, we cross-check the events each segment is waiting for
// RD: against all the events notified in the each other segment

  for(int column=0; column < size_of_conflict_table_; column++) {

    // Here, we are getting the segment id and instance id with the associated
    // column index 'column'.
    std::pair<int, int> column_id_and_instance =
      index_to_segment_and_instance_id(column);
    VertexDescriptor column_vertex
      = segment_id_to_vertex_descriptor(column_id_and_instance.first);
    Segment &ColSegment = segment_graph_.graph_[column_vertex];

    // RD: now we use the list computed above in Step 1
    // RD: and apply the mapping to instances
    MappedVariableList wait_events = determine_mapped_events(
                                            ColSegment.wait_events_,
                                            column_id_and_instance.second,
                                            dynamic_analysis);

    for(int row=0; row < size_of_conflict_table_; row++) {

      conflict_table_[row][column] = false;

      // Here, we are getting the segment id and instance id with the associated
      // row index 'row'.
      std::pair<int, int> row_id_and_instance =
        index_to_segment_and_instance_id(row);
      VertexDescriptor row_vertex
        = segment_id_to_vertex_descriptor(row_id_and_instance.first);
      Segment &RowSegment = segment_graph_.graph_[row_vertex];

      // RD: now we use the list computed above in Step 2
      // RD: and apply the mapping to instances
      MappedVariableList notify_events = determine_mapped_events(
                                            RowSegment.notify_events_,
                                            row_id_and_instance.second,
                                            dynamic_analysis);

      // RD: for each segment pair, we cross-check the notified events
      // RD: against the events waited for

      for(MappedVariableList::iterator
          iter_notify_event  = notify_events.begin();
          iter_notify_event != notify_events.end();
          iter_notify_event++) {

        std::list<MappedVariable>::iterator conflict  // RD: this is a "match", not a "conflict"
          = std::find(
                     wait_events.begin(),
                     wait_events.end(),
                     *iter_notify_event);

        if(conflict != wait_events.end()) {
          // we found a common even in both lists!
          conflict_table_[row][column] = true;
          if(debugging_) {  // record the triggering event
            conflict_info_[row][column].conflict_variables_.insert(
                                            *iter_notify_event);
          } else {
            break; // no need to search further (RD)
          }
        }
        else //then we check in the global events ZC/2019/02/11
        {
          if(global_events.count(iter_notify_event->symbol_) == 0) continue;
          else //the notified event is a global event
          {
            for(MappedVariableList::iterator
              iter_wait_event  = wait_events.begin();
              iter_wait_event != wait_events.end();
              iter_wait_event++)
            {
              if(iter_wait_event->symbol_ == iter_notify_event->symbol_)
              {
                conflict_table_[row][column] = true;
                if(debugging_) {  // record the triggering event
                  conflict_info_[row][column].conflict_variables_.insert(
                                                  *iter_notify_event);
                } else {
                  break; // no need to search further (RD)
                }
              }
            }
          }
        }
      }
    } // rows
  } // cols
}

std::list<risc::sg::MappedVariable>
risc::sg::EventConflictTable::determine_mapped_events(  // RD, 07/13/18
        std::set<SymbolWithPath> &event_list,
        int instance_id,
        bool dynamic_analysis)
{
  std::list<MappedVariable> mapped_list;

  for(std::set<SymbolWithPath>::iterator
          iter_event  = event_list.begin();
          iter_event != event_list.end();
          iter_event++) {
    const SymbolWithPath &event = (*iter_event);
    MappedVariable mapped_event(event.symbol_, instance_id);
    // cannot be a pointer type, because that is prevented with error message
    assert(!isSgPointerType(mapped_event.symbol_->get_type()));
    if (isSgReferenceType(mapped_event.symbol_->get_type())) {
      // if it is a reference type, we use the InstanceTree to get
      // the real mapped variable
      // Example: wait(ref1);
      MappedVariableList tmp_mapped_event;
      if(dynamic_analysis) {
        tmp_mapped_event = get_mapped_variable_dynamically(
                                       mapped_event.symbol_,
                                       instance_id);
      } else {
        tmp_mapped_event = get_mapped_variable_statically(
                                       mapped_event.symbol_,
                                       instance_id);
      }
      mapped_event = tmp_mapped_event.front(); // RD: not sure why this is a list in the first place
    }
    if(tools::is_type_sc_event(mapped_event.symbol_->get_type())) {
      // RD: it's a regular event, possibly inside a channel (RD)
      // TS: if it is a event variable, we can set the symbol straight
      // Example: wait(e1);
      if(tools::is_channel_variable(mapped_event.symbol_)) {
        mapped_event.instance_id_ = get_channel_id(event, instance_id,
                                                      dynamic_analysis);
      }
      mapped_list.push_back(mapped_event);
    } else {
      // RD: otherwise, this must be an event list (of OR or AND type)
      //....assert
      if(!
            (  (tools::is_type_sc_event_and_list(mapped_event.symbol_->get_type()))
             ||(tools::is_type_sc_event_or_list(mapped_event.symbol_->get_type())))
        ) {
        printf("\nASSERTION WARNING(match): Neither sc_event, nor sc_event_list!\n");
      }
      // here we are covering the case of an event list
      // which has been composed in function before_end_of_elaboration
      // Example: wait(event_list);

      std::list<MappedVariable> tmp_wait_events;
      if(dynamic_analysis) {
        tmp_wait_events = get_all_list_events_dynamically(
                                         mapped_event.symbol_,
                                         instance_id);
      } else {
        tmp_wait_events = get_all_list_events_statically(
                                         mapped_event.symbol_,
                                         instance_id,
                                         event.pcp_);
      }
      mapped_list.splice(mapped_list.end(), tmp_wait_events); // concat!
    }
  }
  return(mapped_list);
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
