#include "read_write_analysis.h"

#include <algorithm>
#include "boost/range/algorithm/set_algorithm.hpp"

#include "../internal_representation/variable.h"

void
risc::sa::collect_write_variables(SgExpression *expr,
    std::vector<SgNode*> &write_var)
{
  // This case catches something like int a = 42; or int m = k[33];
  if(isSgAssignInitializer(expr)) {

    if(isSgInitializedName(expr->get_parent())) {

      // we skip references
      if(SageInterface::isReferenceType(isSgInitializedName(
                                            expr->get_parent())->get_type()))
        return;

      write_var.push_back(expr->get_parent());

    } else {
      // it is something like int base[MAX] = {1};
      // there will be another SgAssignInitializer statement.
      return;
    }
  }

  // This cases catches something like int bucket[BASE] = { 0 };
  if(isSgAggregateInitializer(expr)) {
    if(isSgInitializedName(expr->get_parent())) {
      write_var.push_back(expr->get_parent());
    }
  }

  if(isSgPlusPlusOp(expr) || isSgMinusMinusOp(expr)) {

    SgNode *operand = NULL;

    if(isSgPlusPlusOp(expr))
      operand = isSgPlusPlusOp(expr)->get_operand();

    if(isSgMinusMinusOp(expr))
      operand = isSgMinusMinusOp(expr)->get_operand();

    if(isSgPntrArrRefExp(operand)) {

      SgPntrArrRefExp* array_expr = isSgPntrArrRefExp(operand);

      while(isSgPntrArrRefExp(array_expr->get_lhs_operand()))
        array_expr = isSgPntrArrRefExp(array_expr->get_lhs_operand());

      SgArrowExp *arrow_expr = isSgArrowExp(array_expr->get_lhs_operand());
      if(arrow_expr) {

        // in this case the expression considers a member variable
        if(isSgThisExp(arrow_expr->get_lhs_operand())) {

          // in this case we have something like this->array[5][3]++(--);
          write_var.push_back(isSgVarRefExp(arrow_expr->get_rhs_operand()));
        }

      } else {

        // in this case we have something like array[5][3]++(--);
        write_var.push_back(isSgVarRefExp(array_expr->get_lhs_operand()));
      }
    }
    else if(isSgArrowExp(operand)) {

      SgArrowExp *arrow_expr = isSgArrowExp(operand);

      // in this case the expression considers a member variable
      if(isSgThisExp(arrow_expr->get_lhs_operand())) {

        // in this case we have something like this->member++(--);
        write_var.push_back(isSgVarRefExp(arrow_expr->get_rhs_operand()));
      }
    }
    else {

      // in this case we have something like variable++(--);
      write_var.push_back(isSgVarRefExp(operand));
    }
  }

  if(isSgCompoundAssignOp(expr) || isSgAssignOp(expr)) {

    SgNode *left_operand = NULL;

    if(isSgCompoundAssignOp(expr))
      left_operand = isSgCompoundAssignOp(expr)->get_lhs_operand();

    if(isSgAssignOp(expr))
      left_operand = isSgAssignOp(expr)->get_lhs_operand();

    if(isSgPntrArrRefExp(left_operand)) {

      SgPntrArrRefExp *array_expr = isSgPntrArrRefExp(left_operand);

      while(isSgPntrArrRefExp(array_expr->get_lhs_operand()))
        array_expr = isSgPntrArrRefExp(array_expr->get_lhs_operand());

      SgArrowExp *arrow_expr = isSgArrowExp(array_expr->get_lhs_operand());
      if(arrow_expr) {

        if(isSgThisExp(arrow_expr->get_lhs_operand())) {

          // in this case we have something like this->array[5][3] (+)= ... ;
          write_var.push_back(isSgVarRefExp(arrow_expr->get_rhs_operand()));
        }

      } else {

        // in this case we have something like array[5][3] (+)= ... ;
        write_var.push_back(isSgVarRefExp(array_expr->get_lhs_operand()));
      }
    }
    else if(isSgArrowExp(left_operand)) {

      SgArrowExp *arrow_expr = isSgArrowExp(left_operand);

      if(arrow_expr) {

        if(isSgThisExp(arrow_expr->get_lhs_operand())) {

          // in this case we have something like this->member (+)= ...;
          write_var.push_back(isSgVarRefExp(arrow_expr->get_rhs_operand()));
        }
      }
    }
    else {
      // in this case we have something like local (+)= ..;
      write_var.push_back(isSgVarRefExp(left_operand));
    }
  }
}


void
risc::sa::collect_read_variables(SgExpression *expr,
    std::vector<SgNode*> &read_var)
{
  if(isSgExpression(expr->get_parent()))
    return;

  SgExpression *sub_expr_lhs = NULL;
  SgExpression *sub_expr_rhs = NULL;

  if(isSgPlusPlusOp(expr) || isSgMinusMinusOp(expr)) {

    SgNode *operand = NULL;

    if(isSgPlusPlusOp(expr))
      operand = isSgPlusPlusOp(expr)->get_operand();

    if(isSgMinusMinusOp(expr))
      operand = isSgMinusMinusOp(expr)->get_operand();

    // we have array access
    if(isSgPntrArrRefExp(operand)) {

      // this cachtes this->member[3]++(--) and local[3]++(--);
      sub_expr_rhs = isSgPntrArrRefExp(operand);
    }
    else if(isSgArrowExp(operand)) {

      SgArrowExp *arrow_expr = isSgArrowExp(operand);

      // in this case the expression considers a member variable
      if(isSgThisExp(arrow_expr->get_lhs_operand())) {

        // in this case we have something like this->member++(--);
        read_var.push_back(isSgVarRefExp(arrow_expr->get_rhs_operand()));
      }
    } else {

      // in this case we have something like local++(--);
      read_var.push_back(isSgVarRefExp(operand));
    }
  }

  if(isSgAssignInitializer(expr)) {
    sub_expr_rhs = expr;
  }

  if(isSgAggregateInitializer(expr)) {
    sub_expr_rhs = expr;
  }

  if(isSgBinaryOp(expr)) {

    if(isSgCompoundAssignOp(expr)) {
      sub_expr_rhs = isSgCompoundAssignOp(expr)->get_rhs_operand();
      sub_expr_lhs = isSgCompoundAssignOp(expr)->get_lhs_operand();

      if(isSgPntrArrRefExp(isSgCompoundAssignOp(expr)->get_lhs_operand())) {

        // getting all right VarRefExp
        SgPntrArrRefExp* array_expr =
          isSgPntrArrRefExp(isSgCompoundAssignOp(expr)->get_lhs_operand());

        std::vector<SgNode*> temp_var_ref =
          NodeQuery::querySubTree(array_expr->get_rhs_operand(), V_SgVarRefExp);

        read_var.insert(read_var.end(), temp_var_ref.begin(),
                     temp_var_ref.end());


        while(isSgPntrArrRefExp(array_expr->get_lhs_operand())) {

          array_expr = isSgPntrArrRefExp(array_expr->get_lhs_operand());

          std::vector<SgNode*> temp_var_ref =
            NodeQuery::querySubTree(array_expr->get_rhs_operand(),
                           V_SgVarRefExp);
          read_var.insert(read_var.end(), temp_var_ref.begin(),
                       temp_var_ref.end());
        }
      }
    }

    else if(isSgAssignOp(expr)) {
      sub_expr_rhs = isSgAssignOp(expr)->get_rhs_operand();
      sub_expr_lhs = NULL;

      if(isSgPntrArrRefExp(isSgAssignOp(expr)->get_lhs_operand())) {


        SgPntrArrRefExp* array_expr =
          isSgPntrArrRefExp(isSgAssignOp(expr)->get_lhs_operand());

        std::vector<SgNode*> temp_var_ref =
          NodeQuery::querySubTree(array_expr->get_rhs_operand(), V_SgVarRefExp);

        read_var.insert(read_var.end(), temp_var_ref.begin(),
                     temp_var_ref.end());

        while(isSgPntrArrRefExp(array_expr->get_lhs_operand())){

          array_expr = isSgPntrArrRefExp(array_expr->get_lhs_operand());

          std::vector<SgNode*> temp_var_ref =
            NodeQuery::querySubTree(array_expr->get_rhs_operand(),
                           V_SgVarRefExp);
          read_var.insert(read_var.end(), temp_var_ref.begin(),
                       temp_var_ref.end());
        }
      }
    }

    else {
      sub_expr_lhs = expr;
    }
  }

  if(sub_expr_lhs != NULL) {

    std::vector<SgNode*> temp_var_ref =
      NodeQuery::querySubTree(sub_expr_lhs , V_SgVarRefExp);

    read_var.insert(read_var.end(), temp_var_ref.begin(), temp_var_ref.end());
  }

  if(sub_expr_rhs != NULL) {

    std::vector<SgNode*> temp_var_ref =
      NodeQuery::querySubTree(sub_expr_rhs , V_SgVarRefExp);

    read_var.insert(read_var.end(), temp_var_ref.begin(), temp_var_ref.end());
  }
}

// TODO undocumented helper function
bool
symbol_is_in_set(std::set<SgVariableSymbol*> &reference_set, SgNode* node)
{
  SgVariableSymbol* symbol = NULL;

  if(isSgInitializedName(node)) {
    symbol = isSgVariableSymbol(
                 isSgInitializedName(node)->
                     search_for_symbol_from_symbol_table());
  }

  if(isSgVarRefExp(node)) {
    symbol = isSgVarRefExp(node)->get_symbol();
  }

  for(std::set<SgVariableSymbol*>::iterator
      iter  = reference_set.begin();
      iter != reference_set.end(); iter++) {

    if((*iter) == symbol)
      return true;
  }
  return false;
}

// TODO undocumented helper function
std::vector<SgNode*>::iterator
delete_if_variables_have_same_symbol(
    std::vector<SgNode*>::iterator first,
    std::vector<SgNode*>::iterator last,
    std::set<SgVariableSymbol*> &reference_set)
{
  //ForwardIterator result = first;
  std::vector<SgNode*>::iterator result = first;
  while (first!=last) {
    if (!symbol_is_in_set(reference_set, *first)) {
      *result = *first;
      ++result;
    }
    ++first;
  }
  return result;
}


void
risc::sa::read_write_analysis(
    std::vector<SgExpression*> expressions,
    std::set<SgVariableSymbol*> &read_only_var,
    std::set<SgVariableSymbol*> &write_only_var,
    std::vector<SgNode*> &read_write)
{
  std::vector<SgNode*> temp_read_node, temp_write_node;

  for(std::vector<SgExpression*>::iterator
      iter = expressions.begin(); iter != expressions.end(); iter++) {

    collect_read_variables(*iter, temp_read_node);
    collect_write_variables(*iter, temp_write_node);
  }

  std::set<SgVariableSymbol*> temp_read, temp_write;

  // collecting all reading symbols
  for(std::vector<SgNode*>::iterator
      iter  = temp_read_node.begin();
      iter != temp_read_node.end();
      iter++) {

    if(isSgVarRefExp(*iter)) {

      temp_read.insert(isSgVarRefExp(*iter)->get_symbol());
    }

    if(isSgInitializedName(*iter)) {

      temp_read.insert(
                    isSgVariableSymbol(
                        isSgInitializedName(*iter)->
                            search_for_symbol_from_symbol_table()));
    }
  }

  // collecting all writing symbols
  for(std::vector<SgNode*>::iterator
      iter  = temp_write_node.begin();
      iter != temp_write_node.end();
      iter++) {

    if(isSgVarRefExp(*iter)) {

      temp_write.insert(isSgVarRefExp(*iter)->get_symbol());
    }

    if(isSgInitializedName(*iter)) {

      temp_write.insert(
                     isSgVariableSymbol(
                         isSgInitializedName(*iter)->
                             search_for_symbol_from_symbol_table()));
    }
  }

  // compute the write only symbols
  boost::set_difference(
             temp_write,
             temp_read,
             std::inserter(
                      write_only_var,
                      write_only_var.begin()));

  // compute the read only symbols
  boost::set_difference(
             temp_read,
             temp_write,
             std::inserter(
                      read_only_var,
                      read_only_var.begin()));


  read_write = temp_read_node;
  read_write.insert(
                 read_write.end(),
                 temp_write_node.begin(),
                 temp_write_node.end());

  // delete the read only variables from the read/write vector
  read_write.erase(
                 delete_if_variables_have_same_symbol(
                     read_write.begin(),
                     read_write.end(),
                     read_only_var),
                 read_write.end());

  // delete the write only variables from the read/write vector
  read_write.erase(
                 delete_if_variables_have_same_symbol(
                     read_write.begin(),
                     read_write.end(),
                     write_only_var),
                 read_write.end());
}

// TODO undocumented helper function
bool
risc::sa::is_member_variable_symbol(SgNode *given_variable,
    VariableVector &member_variables)
{
  // Iterating all the member variables
  SgVariableSymbol *symbol = NULL;

  if(isSgInitializedName(given_variable)) {
    symbol = isSgVariableSymbol(
                 isSgInitializedName(given_variable)->
                     search_for_symbol_from_symbol_table());
  } else if(isSgVarRefExp(given_variable)) {
    symbol = isSgVarRefExp(given_variable)->get_symbol();
  } else if(isSgVariableSymbol(given_variable)) {
    symbol = isSgVariableSymbol(given_variable);
  } else {
    std::cout << "Not covered situation" << std::endl;
  }

  // iterating over member variables
  for(VariableVectorConstIter
      iter  = member_variables.begin();
      iter != member_variables.end();
      iter++) {

    SgVariableDefinition *var_def = (*iter)->get_ast_node();

    // for some unknown reason we have to get the symbol over this complicated
    // way.
    SgVariableSymbol *member_symbol =
      isSgVariableSymbol(
          (*iter)->get_ast_node()->get_declaration()->
              get_decl_item(0)->search_for_symbol_from_symbol_table());

    if(member_symbol == symbol) {

      return true;
    }
  }
  return false;
}

// TODO undocumented helper function
std::vector<SgNode*>::iterator
risc::sa::delete_if_is_not_member_variable(
    std::vector<SgNode*> &given_variables, VariableVector &member_variables)
{
  std::vector<SgNode*>::iterator first = given_variables.begin();
  std::vector<SgNode*>::iterator last = given_variables.end();

  // Iterating all the given variables
  std::vector<SgNode*>::iterator result = first;
  while(first!=last) {
    if(is_member_variable_symbol(*first, member_variables)) {
      *result = *first;
      ++result;
    }
    ++first;
  }
  return result;
}

// TODO undocumented helper function
std::set<SgVariableSymbol*>::iterator
risc::sa::delete_if_is_not_member_symbol(
    std::set<SgVariableSymbol*> &given_variables,
    VariableVector &member_variables)
{
  std::set<SgVariableSymbol*>::iterator first = given_variables.begin();
  std::set<SgVariableSymbol*>::iterator last = given_variables.end();

  // Iterating all the given variables
  std::set<SgVariableSymbol*>::iterator result = first;
  while(first!=last) {
    if(!is_member_variable_symbol(*first, member_variables)) {
      given_variables.erase(first++);
    } else {
      ++first;
    }
  }
  return result;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
