// play.cpp: conceptual audio/video stream decoding example
// Author: RISC project members, CECS
// 
// 27/11/18 ZC  use sockets to replace ports and channels
// 06/05/18 RD	confirmed clean sc_time constants with GCC 6.3.1
// 04/27/18 RD  clean sc_time constants for newer compilers (e.g. GCC 4.9.2)
// 04/26/18 RD  merged tracing version into main file (define TRACE_FILE)
// 04/24/18 DM  added tracing of stream data (sc_trace is now supported)
// 07/10/17 RD  split the delay into two parts (as suggested by ZC)
// 04/05/17 RD  removed sc_stop from stimulus, let monitor terminate;
//              added COUNT_FRAMES to ensure proper completion
// 03/08/17 RD  avoid duplicated DELAY macros, set BUFSIZE to 10
// 01/13/17 GL  separate version play_fifo.cpp with sc_fifo channels
// 10/04/16 RD  removed obsolete C artifacts from the user-defined queue channel
// 09/30/16 RD  cleanup for release 0.3.0
// 03/07/15 RD  bug fix: use sc_stop() to quit cleanly (exit() is not MT-safe)
// 10/30/15 RD  polished for release 0.2.1
// 09/04/15 GL  use 'long long' (not 'double') to avoid floating-point overflow
//              (FPU exception handling shows high performance variation;
//              moreover, integer arithmetic is more realistic for decoder)
// 08/30/15 GL  initial version based on SpecC example play.sc by RD


#include <stdio.h>
#include "systemc.h"
#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"
#include <string>
using namespace sc_core;
using namespace std;

// frame rates
#define AUDIO_DELAY     26120, SC_US
#define VIDEO_DELAY     33330, SC_US
#define HALF_AUDIO_DELAY 26120/2, SC_US
#define HALF_VIDEO_DELAY 33330/2, SC_US

// simulated time
#define CLIP_LENGTH     10, SC_SEC

// simulated workload unit
#define AUDIOLOAD        7500000
#define VIDEOLOAD       10000000

// illustrate and check frame decoding
#define PRINT_TIME
#define PRINT_FRAMES
#define COUNT_FRAMES
//#define TRACE_FILE

// buffer size in queue channels
#define BUFSIZE		10ul

SC_MODULE(Stimulus)
{
  // sc_fifo_out<long long> VideoStream;
  // sc_fifo_out<long long> AudioStreamL;
  // sc_fifo_out<long long> AudioStreamR;
  tlm_utils::simple_initiator_socket<Stimulus> VideoStream;
  tlm_utils::simple_initiator_socket<Stimulus> AudioStreamL;
  tlm_utils::simple_initiator_socket<Stimulus> AudioStreamR;
  long long video;
  long long audioLeft;
  long long audioRight;


  void Video()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time VideoDelay(VIDEO_DELAY);
    sc_time t1(0, SC_US);
    while(t1 < ClipLength){
      t1 += VideoDelay;
      tlm::tlm_command cmd = tlm::TLM_WRITE_COMMAND;
      sc_time delay0 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload trans_stim_v; 
      trans_stim_v.set_command( cmd );
      trans_stim_v.set_data_ptr( reinterpret_cast<unsigned char*>(&video) );
      trans_stim_v.set_data_length( 8 );
      VideoStream->b_transport(trans_stim_v, delay0);
    }
  }

  void AudioL()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);
    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      tlm::tlm_command cmd = tlm::TLM_WRITE_COMMAND;
      sc_time delay1 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload trans_stim_al; 
      trans_stim_al.set_command( cmd );
      trans_stim_al.set_data_ptr( reinterpret_cast<unsigned char*>(&audioLeft) );
      trans_stim_al.set_data_length( 8 );
      AudioStreamL->b_transport(trans_stim_al, delay1);
    }
  }

  void AudioR()
  {

    sc_time ClipLength(CLIP_LENGTH);
    sc_time AudioDelay(AUDIO_DELAY);
    sc_time t2(0, SC_US);
    while(t2 < ClipLength)
    {
      t2 += AudioDelay;
      tlm::tlm_command cmd = tlm::TLM_WRITE_COMMAND;
      sc_time delay2 = sc_time(0, SC_NS);
      tlm::tlm_generic_payload trans_stim_ar; 
      trans_stim_ar.set_command( cmd );
      trans_stim_ar.set_data_ptr( reinterpret_cast<unsigned char*>(&audioRight) );
      trans_stim_ar.set_data_length( 8 );
      AudioStreamR->b_transport(trans_stim_ar, delay2);
    }
  }
  
  SC_CTOR(Stimulus)
  {
    video = 3;
    audioLeft = 27;
    audioRight = 15;
    SC_THREAD(Video);
    SC_THREAD(AudioL);
    SC_THREAD(AudioR);
  }
};

SC_MODULE(VideoCodec)
{
  tlm_utils::simple_target_socket<VideoCodec>  p1;
  tlm_utils::simple_initiator_socket<VideoCodec> p2;

  long long d, outFrame, inFrame;
  sc_event e_p1, e_p2;
  
  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }
  
  void receive(tlm::tlm_generic_payload& trans_bt_v, sc_time& delay)
  {
    tlm::tlm_command cmd = trans_bt_v.get_command();
    unsigned char*   ptr_v = trans_bt_v.get_data_ptr();
    unsigned int     len = trans_bt_v.get_data_length();
    if ( cmd == tlm::TLM_WRITE_COMMAND )
      memcpy(&inFrame, ptr_v, len);
    else{
      std::cout << "wrong command" << std::endl;
      exit(0);
    }

    e_p1.notify(SC_ZERO_TIME);
    wait(e_p2);
  }
  
  void main(void)
  {

    d = 1;

    while(1)
    {
      wait(e_p1);
      wait(HALF_VIDEO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_VIDEO_DELAY);

      tlm::tlm_generic_payload trans_v; 
      sc_time delay = sc_time(0, SC_NS);
      tlm::tlm_command cmd = tlm::TLM_WRITE_COMMAND;
      trans_v.set_command( cmd );
      trans_v.set_data_ptr( reinterpret_cast<unsigned char*>(&outFrame) );
      trans_v.set_data_length( 8 );
      p2->b_transport(trans_v, delay);
      e_p2.notify(SC_ZERO_TIME);
    }
  }

  SC_CTOR(VideoCodec)
  {
    p1.register_b_transport(this, &VideoCodec::receive);
    SC_THREAD(main);
  }
};

SC_MODULE(AudioCodec)
{
  tlm_utils::simple_target_socket<AudioCodec>  p1;
  tlm_utils::simple_initiator_socket<AudioCodec> p2;
  sc_event e_p1, e_p2;
  long long d, inFrame, outFrame;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }
  
  
  void receive(tlm::tlm_generic_payload& trans_bt_a, sc_time& delay)
  {
    tlm::tlm_command cmd = trans_bt_a.get_command();
    unsigned char*   ptr_a = trans_bt_a.get_data_ptr();
    unsigned int     len = trans_bt_a.get_data_length();
    if ( cmd == tlm::TLM_WRITE_COMMAND )
      memcpy(&inFrame, ptr_a, len);
    else{
      std::cout << "wrong command" << std::endl;
      exit(0);
    }
    e_p1.notify(SC_ZERO_TIME);
    wait(e_p2);
  }
  
  void main(void)
  {

    d = 1;

    while(1)
    {
      wait(e_p1);
      wait(HALF_AUDIO_DELAY);
      outFrame = decode(inFrame);
      wait(HALF_AUDIO_DELAY);
     
      tlm::tlm_generic_payload trans_a; 
      sc_time delay = sc_time(0, SC_NS);
      tlm::tlm_command cmd = tlm::TLM_WRITE_COMMAND;
      trans_a.set_command( cmd );
      trans_a.set_data_ptr( reinterpret_cast<unsigned char*>(&outFrame) );
      trans_a.set_data_length( 8 );
      p2->b_transport(trans_a, delay);
      e_p2.notify(SC_ZERO_TIME);
    }
  }

  SC_CTOR(AudioCodec)
  {
    p1.register_b_transport(this, &AudioCodec::receive);
    SC_THREAD(main);
  }
};

SC_MODULE(DUT)
{
  VideoCodec video;
  AudioCodec left, right;
  SC_CTOR(DUT): video("video"), left("left"), right("right")
  { }
};

SC_MODULE(Display)
{
  tlm_utils::simple_target_socket<Display> VideoIn;
  sc_event e_VideoIn, e_p2;
  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesDisplayed;
#endif

  void display(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar('V');       // display the frame
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesDisplayed++;
#endif
  }
  
  void receive(tlm::tlm_generic_payload& trans_bt_d, sc_time& delay)
  {
    tlm::tlm_command cmd = trans_bt_d.get_command();
    unsigned char*   ptr_d = trans_bt_d.get_data_ptr();
    unsigned int     len = trans_bt_d.get_data_length();
    if ( cmd == tlm::TLM_WRITE_COMMAND )
      memcpy(ptr_d, &Frame, len);
    else{
      std::cout << "wrong command" << std::endl;
      exit(0);
    }
    e_VideoIn.notify(SC_ZERO_TIME);
    wait(e_p2);
  }
  
  void main(void)
  {
    
    while(1)
    {
      wait(e_VideoIn);
      display(Frame);
      e_p2.notify(SC_ZERO_TIME);
    }
  }

  SC_CTOR(Display)
  {
#ifdef COUNT_FRAMES
    FramesDisplayed = 0;
#endif
    VideoIn.register_b_transport(this, &Display::receive);
    SC_THREAD(main);
  }
  
};

SC_MODULE(Speaker)
{
  tlm_utils::simple_target_socket<Speaker> AudioIn;
  sc_event e_AudioIn, e_p2;
  long long Frame;
#ifdef COUNT_FRAMES
  unsigned FramesPlayed;
#endif

  char Channel;

  void sound(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar(Channel); // output the sound
    fflush(stdout);
#endif
#ifdef COUNT_FRAMES
    FramesPlayed++;
#endif
  }

  void receive(tlm::tlm_generic_payload& trans_bt_s, sc_time& delay)
  {
    tlm::tlm_command cmd = trans_bt_s.get_command();
    unsigned char*   ptr_s = trans_bt_s.get_data_ptr();
    unsigned int     len = trans_bt_s.get_data_length();
    if ( cmd == tlm::TLM_WRITE_COMMAND )
      memcpy(&Frame, ptr_s, len);
    else{
      std::cout << "wrong command" << std::endl;
      exit(0);
    }
    e_AudioIn.notify(SC_ZERO_TIME);
    wait(e_p2);
  }
  
  
  void main(void)
  {
    long long Frame;

    while(1)
    {
      wait(e_AudioIn);
      sound(Frame);
      e_p2.notify(SC_ZERO_TIME);
    }
  }

  Speaker(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
    
#ifdef COUNT_FRAMES
    FramesPlayed = 0;
#endif
    AudioIn.register_b_transport(this, &Speaker::receive);
    SC_THREAD(main);
  }
  SC_HAS_PROCESS(Speaker);
};

SC_MODULE(Monitor)
{
  Display screen;
  Speaker speakerL, speakerR;

  SC_CTOR(Monitor):
        screen("screen"),
        speakerL("speakerL", 'L'),
        speakerR("speakerR", 'R')
  { }
};

SC_MODULE(Top)
{
  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;

  void before_end_of_elaboration()
  {
    stimulus.VideoStream.bind(dut.video.p1);
    stimulus.AudioStreamL.bind(dut.left.p1);
    stimulus.AudioStreamR.bind(dut.right.p1);

    dut.video.p2.bind(monitor.screen.VideoIn);
    dut.left.p2.bind(monitor.speakerL.AudioIn);
    dut.right.p2.bind(monitor.speakerR.AudioIn);
    
  }

  SC_CTOR(Top)
  : stimulus("stimulus")
  , dut("dut")
  , monitor("monitor")
  { }
};

int sc_main(int argc, char* argv[])
{
  Top top("top");
  sc_time	ClipLength(CLIP_LENGTH);
  sc_time	VideoDelay(VIDEO_DELAY);
  sc_time	AudioDelay(AUDIO_DELAY);
#ifdef TRACE_FILE
  sc_trace_file *f = sc_create_vcd_trace_file(argv[0]);
  f->set_time_unit(sc_time(10, SC_US).to_seconds(), SC_SEC);
  sc_trace(f, top.dut.video.d, "top.dut.video.d");
  sc_trace(f, top.dut.left.d, "top.dut.left.d");
  sc_trace(f, top.dut.right.d, "top.dut.right.d");
#  ifdef COUNT_FRAMES
  sc_trace(f, top.monitor.screen.FramesDisplayed, "top.monitor.screen.FramesDisplayed");
  sc_trace(f, top.monitor.speakerL.FramesPlayed, "top.monitor.speakerL.FramesPlayed");
  sc_trace(f, top.monitor.speakerR.FramesPlayed, "top.monitor.speakerR.FramesPlayed");
#  endif
#endif

#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
#endif
  printf("Playing...\n");
  sc_start();
#ifdef PRINT_FRAMES
  printf("\n");
#endif
#ifdef COUNT_FRAMES
  printf("%u video frames displayed\n", top.monitor.screen.FramesDisplayed);
  printf("%u left audio frames played\n", top.monitor.speakerL.FramesPlayed);
  printf("%u right audio frames played\n", top.monitor.speakerR.FramesPlayed);
  assert(top.monitor.screen.FramesDisplayed ==
		ClipLength.value()/VideoDelay.value()+1);
  assert(top.monitor.speakerL.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
  assert(top.monitor.speakerR.FramesPlayed ==
		ClipLength.value()/AudioDelay.value()+1);
#endif
#ifdef PRINT_TIME
  printf("%s: ", sc_time_stamp().to_string().c_str());
  assert(sc_time_stamp() >= ClipLength);
#endif
#ifdef TRACE_FILE
  sc_close_vcd_trace_file(f);
#endif
  printf("Done!\n");
  return 0;
}

// EOF
