// bug20181129.cc: WAS: mandelbrot_crunch.cc: Mandelbrot set renderer example in SystemC
// bug fix for incorrect HTML conflict tables (i.e. "mc_data_conflict_table.html")
// 11/30/18, RD

#include <stdio.h>
#include "systemc.h"


// Definitions ///////////////////////////////////////////////////////////


#define WIDTH           640
#define HEIGHT          512
#define NUM_IMAGES      20


// Data structures ///////////////////////////////////////////////////////


typedef struct Image
{
#pragma RISC R conflict-free
    unsigned char R[HEIGHT][WIDTH];
#pragma RISC G conflict-free
    unsigned char G[HEIGHT][WIDTH];
#pragma RISC B conflict-free
    unsigned char B[HEIGHT][WIDTH];

    Image(void)
    {
        for (int i=0; i<HEIGHT; i++)
            for (int j=0; j<WIDTH; j++)
            {
                R[i][j] = 0;
                G[i][j] = 0;
                B[i][j] = 0;
            }
    }

    Image& operator=(const Image& copy)
    {
        for (int i=0; i<HEIGHT; i++)
            for (int j=0; j<WIDTH; j++)
            {
                R[i][j] = copy.R[i][j];
                G[i][j] = copy.G[i][j];
                B[i][j] = copy.B[i][j];
            }
        return *this;
    }

    friend ostream& operator<<(ostream& os, const Image& im);
} IMAGE;

ostream& operator<<(ostream& os, const Image& im)
{
    for (int i=0; i<HEIGHT; i++)
        for (int j=0; j<WIDTH; j++)
        {
            os << im.R[i][j] << ' ';
            os << im.G[i][j] << ' ';
            os << im.B[i][j] << '\n';
        }
    return os;
}


// Channels //////////////////////////////////////////////////////////////


// user-defined FIFO channels

class i_image_sender: virtual public sc_interface
{
    public:
        virtual void write(IMAGE&) = 0;
};

class i_image_receiver: virtual public sc_interface
{
    public:
        virtual void read(IMAGE&) = 0;
};

class c_image_channel: public sc_channel, public i_image_sender, public i_image_receiver
{
    public:
        c_image_channel(sc_module_name name, int size_ = 4): sc_channel(name)
        {
            size = size_;
            buf = new IMAGE[size];
            reset();
        }

        ~c_image_channel()
        {
            delete [] buf;
        }

        int num_available() const
        {
            return size - free_slots;
        }

        int num_free() const
        {
            return free_slots;
        }

        void write(IMAGE& X)
        {
            while (num_free() == 0)
                wait(data_read_event);
            buf[wi] = X;
            wi = (wi+1) % size;
            free_slots --;
            data_written_event.notify(SC_ZERO_TIME);
        }

        void read(IMAGE& Y)
        {
            while (num_available() == 0)
                wait(data_written_event);
            Y = buf[ri];
            ri = (ri+1) % size;
            free_slots ++;
            data_read_event.notify(SC_ZERO_TIME);
        }

        void reset()
        {
            free_slots = size;
            ri = 0;
            wi = 0;
        }

    private:
        int size;       // size
        IMAGE *buf;     // circular buffer
        int free_slots; // free space
        int ri;         // read index
        int wi;         // write index

        sc_event data_read_event;
        sc_event data_written_event;
};


// Modules /////////////////////////////////////////////////////////////


SC_MODULE(Monitor)
{
    sc_port<i_image_receiver> ImgIn;
    IMAGE image;        // current received image

    int SaveImage(const char *Filename,
                unsigned char R[HEIGHT][WIDTH],
                unsigned char G[HEIGHT][WIDTH],
                unsigned char B[HEIGHT][WIDTH])
    {
    FILE    *File;
    int     y, x;

    assert(Filename);
    assert(strlen(Filename) < 80);
    File = fopen(Filename, "w");
    if (!File)
    {
        printf("\nCannot open file \"%s\" for writing!\n", Filename);
        return 1;
    }
    fprintf(File, "P6\n");
    fprintf(File, "%d %d\n", WIDTH, HEIGHT);
    fprintf(File, "255\n");
    for(y=0; y<HEIGHT; y++)
    {
        for(x=0; x<WIDTH; x++)
        {
            fputc(R[y][x], File);
            fputc(G[y][x], File);
            fputc(B[y][x], File);
        }
    }
    if (ferror(File))
    {
        printf("\nFile error while writing to file!\n");
        fclose(File);
        return 2;
    }
    fclose(File);
    return 0;
    } /* end of SaveImage */

    void main(void)
    {
        int i;

        for(i=0; i<NUM_IMAGES; i++)
        {
            ImgIn->read(image);
        }
        sc_stop();
    }

    SC_CTOR(Monitor)
    {
        SC_THREAD(main);
    }
};


SC_MODULE(DataOut)
{
    sc_port<i_image_receiver> ImgIn;
    sc_port<i_image_sender> ImgOut;

    IMAGE image;

    void main(void)
    {
        while(1)
        {
            ImgIn->read(image);
            ImgOut->write(image);
        }
    }

    SC_CTOR(DataOut)
    {
        SC_THREAD(main);
    }
};


SC_MODULE(Platform)
{
//    sc_port<i_coords_receiver> CoordsIn;
    sc_port<i_image_sender> ImgOut;

//    c_coords_channel   platform_c1;
    c_image_channel    platform_c2;
//    DataIn          din;
//    DUT             dut;
    DataOut         dout;

    SC_CTOR(Platform):
//        platform_c1("platform_c1"),
        platform_c2("platform_c2"),
//        din("din"),
//        dut("dut"),
        dout("dout")
    {
    }

    void before_end_of_elaboration()
    {
//        din.CoordsIn.bind(CoordsIn);
//        din.CoordsOut.bind(platform_c1);

//        dut.CoordsIn.bind(platform_c1);
//        dut.ImgOut.bind(platform_c2);

        dout.ImgIn.bind(platform_c2);
        dout.ImgOut.bind(ImgOut);
    }
};


SC_MODULE(TOP)
{
//    c_coords_channel   c1;
    c_image_channel    c2;

 //   Stimulus    stimulus;
    Platform    platform;
    Monitor     monitor;

    SC_CTOR(TOP):
//        c1("c1"),
        c2("c2"),
//        stimulus("stimulus"),
        platform("platform"),
        monitor("monitor")
    {
    }

    void before_end_of_elaboration()
    {
//        stimulus.CoordsOut.bind(c1);

//        platform.CoordsIn.bind(c1);
        platform.ImgOut.bind(c2);

        monitor.ImgIn.bind(c2);
    }
};


int sc_main(int argc, char* argv[])
{
    TOP top("top");
    sc_start();
    return 0;
}

// EOF
