#include <iostream>

#include "risc.h"

void print_declarations(SgFunctionDefinition *func_def)
{
  std::vector<SgNode*> var_decls
    = NodeQuery::querySubTree(func_def->get_body(), V_SgVariableDeclaration);

  if(var_decls.begin() != var_decls.end()) {
    std::cout << "     decls:" << std::endl;
  }

  for(std::vector<SgNode*>::iterator
      iter  = var_decls.begin();
      iter != var_decls.end();
      iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
    SgVariableSymbol *symbol = isSgVariableSymbol(var_decl->get_decl_item(0)->search_for_symbol_from_symbol_table());

    if(isSgReferenceType(symbol->get_type())) {

      risc::sg::LocalReferenceAttribute *attribute
        = dynamic_cast<risc::sg::LocalReferenceAttribute*>(symbol->getAttribute("local_reference"));

      if(attribute) {
        std::cout << "      " << var_decl->unparseToString() << "(" << symbol << ")" << " -> " << attribute->mapped_variable_ << std::endl;
      }
    } else {
      std::cout << "      " << var_decl->unparseToString() << "(" << symbol << ")" << std::endl;
    }
  }
}

void print_parameter_declarations(SgFunctionDeclaration *func_decl)
{
  if(func_decl->get_args().begin() != func_decl->get_args().end()) {
    std::cout << "     params:" << std::endl;
  }

  for(SgInitializedNamePtrList::iterator
      iter  = func_decl->get_args().begin();
      iter != func_decl->get_args().end();
      iter++) {

    SgVariableSymbol *symbol = isSgVariableSymbol((*iter)->search_for_symbol_from_symbol_table());
    std::cout << "      " << (*iter)->unparseToString() << "(" << symbol << ")" <<  " -> ";

    risc::sg::FunctionParameterReferenceAttribute *reference_annotation
      = dynamic_cast<risc::sg::FunctionParameterReferenceAttribute*>(symbol->getAttribute("references"));

    if(reference_annotation) {

      for(std::set<SgVariableSymbol*>::iterator
          symbol_iter  = reference_annotation->mapped_variables_.begin();
          symbol_iter != reference_annotation->mapped_variables_.end();
          symbol_iter++) {
        std::cout << (*symbol_iter)->get_declaration()->unparseToString() << ", ";
      }

    }
    std::cout << std::endl;
  }

  //if(func_decl->get_args().begin() != func_decl->get_args().end()) {
  //  std::cout << std::endl;
  //}
}

int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);

  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "  " << (*iter_thread)->get_name() << std::endl;
      sg.add_function(*iter_thread, *iter_thread);
    }
    std::cout << std::endl;
  }

  sg.clean_graph();

  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    if((*iter_module)->get_name() != "Top") {
      continue;
    }

    std::cout << "Module: " <<  (*iter_module)->get_name() << std::endl;

#if 0
    std::cout << "  Threads" << std::endl;
    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "   " << (*iter_thread)->get_name() << std::endl;

      SgFunctionDefinition *func_def = (*iter_thread)->get_ast_node();
      SgFunctionDeclaration *temp = isSgFunctionDeclaration(func_def->get_declaration()->get_definingDeclaration());

      if(SageInterface::get_name(temp) == "thread6" || SageInterface::get_name(temp) == "thread7") {
        print_parameter_declarations(temp);
        print_declarations(func_def);
      }
    }
#endif

#if 0
    std::cout << "  Functions" << std::endl;
    for(risc::FunctionVectorIter
        func_iter  = (*iter_module)->functions_.begin();
        func_iter != (*iter_module)->functions_.end();
        func_iter++) {

      std::cout << "    " << (*func_iter)->get_name() << std::endl;

      SgFunctionDefinition *func_def = (*func_iter)->get_ast_node();
      SgFunctionDeclaration *temp = isSgFunctionDeclaration(func_def->get_declaration()->get_definingDeclaration());

      print_parameter_declarations(temp);
      print_declarations(func_def);
    }
#endif
  }

  std::cout << "++++++++++++++ Generation of the DataConflictTable ++++++++++++++" << std::endl;

  risc::sg::DataConflictTable dct(sg, d.path_instance_mapper_, true);

  if(d.perform_static_analysis()) {
    sg.print_graph("references_static.dot");
    sg.print_graph_read_write_access("references_read_write_static.dot");
    dct.print_conflict_table("references_static.html");

  } else {

    sg.print_graph("references_dynamic.dot");
    sg.print_graph_read_write_access("references_read_write_dynamic.dot");
    dct.print_conflict_table("references_dynamic.html");
  }

  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
